// Copyright (C) 2019  Wilko Manger
// Copyright (C) 2019  Mathieu Velten
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:convert';

import 'package:kiwi/kiwi.dart';
import 'package:matrix_sdk/src/api/client.dart';
import 'package:matrix_sdk/src/event/event.dart';
import 'package:matrix_sdk/src/event/room/state/member_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_creation_event.dart';
import 'package:matrix_sdk/src/event/timeline.dart';
import 'package:matrix_sdk/src/event/room/redaction_event.dart';
import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_avatar_change_event.dart';
import 'package:matrix_sdk/src/event/room/state/room_name_change_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/members.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:chopper/chopper.dart';
import 'package:matrix_sdk/src/util/random.dart';
import 'package:meta/meta.dart';
import 'package:url/url.dart';

class Room extends Identifiable<RoomId> {
  @override
  final RoomId id;

  Timeline _timeline;
  Timeline get timeline => _timeline;
  final Members members;

  //// @internal
  String previousBatch;

  final String _scope;
  ClientService _client;

  User _creator;
  /// [UserId] of the user that created this room.
  User get creator => _creator;

  final LocalUser _me;

  Room({
    @required this.id,
    @required String scope,
    String name,
    Url avatarUrl,
    User directUser,
    String previousBatch,
    User creator
  }) : _scope = scope,
       _name = name,
       _avatarUrl = avatarUrl,
       _directUser = directUser,
       _creator = creator,
       _me = Container().resolve(scope),
       members = Members(id, store: Container().resolve(scope)) {

    _timeline = Timeline(this, store: Container().resolve(scope));
    _client = Container().resolve(_scope);
  }

  String _name;
  String get name => _name;

  Url _avatarUrl;
  Url get avatarUrl => _avatarUrl;

  bool get isDirect => directUser != null;

  User _directUser;
  /// The other user of the direct chat. Is null if [isDirect] is false.
  User get directUser => _directUser;

  /// @internal
  Stream<RoomEvent> retrieveRemoteEvents(int count) async* {
    final response = await _client.roomMessages(
      roomId: id.toString(),
      limit: count,
      from: previousBatch,
    );

    final body = json.decode(response.body);

    previousBatch = body['end'];

    final context = ProcessContext();

    // TODO: When most event types are supported, don't check if they're
    // null anymore (null is now returned if the event type is unsupported)
    final processed = (await _processEvents(body['chunk'].reversed, context))
      .where((event) => event != null);

    for (final event in processed) {
      yield event;
    }
  }

  /// @internal
  Future<void> processSync(Map<String, dynamic> response) async {
    // Only set previous batch from sync when none is set
    if (previousBatch == null) {
      previousBatch = response['timeline']['prev_batch'];
    }

    final context = ProcessContext();

    await _processEvents(response['state']['events'], context);

    final processed = await _processEvents(response['timeline']['events'], context);
    await timeline.addAll(processed);
  }

  Future<List<RoomEvent>> _processEvents(
      Iterable<dynamic> rawEvents, ProcessContext context) async {

    RoomCreationEvent creationEvent;
    final processedEvents = List<RoomEvent>();
    for (final rawEvent in rawEvents) {
      final event = await RoomEvent.fromJson(
        RoomEventArgs(room: this),
        rawEvent,
        context: context
      );

      processedEvents.add(event);

      if (event is RedactionEvent) {
        final redactedId = event.redacts;

        final original = await timeline[redactedId];
        if (original != null && original is! RedactedEvent) {
          await timeline.remove(redactedId);

          await timeline.addAll([
            RedactedEvent(
              redaction: event,
              original: original
            )
          ]);
        }
      } else if (event is RoomNameChangeEvent) {
        var name = event.content.name;
        // Set to null if name is empty
        if (name != null && name.isEmpty) {
          name = null;
        }

        _name = name;
      } else if (event is RoomAvatarChangeEvent) {
        _avatarUrl = event.content.url;
      } else if (event is RoomCreationEvent) {
        creationEvent = event;
      } else if (event is InviteEvent && event.content.isDirect) {
        if (!event.content.subject.isIdenticalTo(_me)) {
          _directUser = event.content.subject;
        } else if (!event.sender.isIdenticalTo(_me)) {
          _directUser = event.sender;
        }
      }
    }

    await members.addAll(context.users);

    // Do this after the members have been added
    // so the correct user state can be set
    if (creationEvent != null) {
      _creator = await members.get(
          creationEvent.content.creatorId,
          at: creationEvent.time
      );

      if (_creator == null) {
        _creator = User(
          id: creationEvent.content.creatorId,
          state: UserState(
            roomId: id,
            since: creationEvent.time
          )
        );

        await members.addAll([_creator]);
      }
    }

    return processedEvents;
  }


  Future<void> processUserAccountData(Map<String, dynamic> raw) async {
    final events = raw['events'] as List<dynamic>;

    if (_directUser == null) {
      final event = events.firstWhere(
              (event) => event['type'] == 'm.direct',
          orElse: () => null
      );

      if (event != null) {
        final content = event['content'] as Map<String, dynamic>;

        await content.forEach((userId, roomIds) async {

          if (UserId.isValidFullyQualified(userId)
              && roomIds.contains(id.toString())) {
            _directUser = await members[UserId(userId)];
          }
        });
      }
    }
  }
}

class ProcessContext {
  final Map<EventId, RoomEvent> timeline = Map();

  /// User which is currently the correct one.
  final Map<UserId, User> currentUsers = Map();

  /// All user states.
  final List<User> users = List();
  
  void setUser(User user) {
    users.add(user);
    currentUsers[user.id] = user;
  }
}

class JoinedRoom extends Room {
  JoinedRoom({
    @required RoomId id,
    @required String scope,
    String name,
    Url avatarUrl,
    User directUser,
    User creator,
    String previousBatch
  }) : super(
    id: id,
    scope: scope,
    name: name,
    avatarUrl: avatarUrl,
    directUser: directUser,
    previousBatch: previousBatch,
    creator: creator
  );

  /// Send an [Event] to the [Room].
  /// Returns the up to date [SentState] of the sent event. (Not
  /// including [SentState.received].
  Stream<SentState> send(EventContent content, {String transactionId}) async* {
    if (transactionId == null) {
      transactionId = randomString();
    }

    final eventArgs = RoomEventArgs(
      id: EventId(transactionId),
      room: this,
      time: DateTime.now(),
      sender: User(
        id: Container().resolve<LocalUser>(_scope).id,
      ),
      unsigned: UnsignedData(transactionId: transactionId),
      sentState: SentState.unsent,
    );
    final event = Event.fromContent(content, eventArgs);

    await timeline.addAll([event]);
    yield SentState.unsent;

    var sent = false;
    while (!sent) {
      Response response;

      try {
        response = await _client.send(
          roomId: id.toString(),
          eventType: content.matrixType,
          txnId: transactionId,
          content: json.encode(content.toJson())
        );

        final body = json.decode(response.body);
        final eventId = EventId(body['event_id']);

        final sentEvent = Event.fromContent(
          content,
          eventArgs.copyWith(
            id: eventId,
            sentState: SentState.sent
          )
        );

        sent = true;
        await timeline.addAll([sentEvent]);
        yield SentState.sent;
      } on Exception {
        yield SentState.failed;
        continue;
      }
    }
  }
}

class LeftRoom extends Room {
  LeftRoom({
    @required RoomId id,
    @required String scope,
    String name,
    Url avatarUrl,
    User directUser,
    User creator,
    String previousBatch
  }) : super(
    id: id,
    scope: scope,
    name: name,
    avatarUrl: avatarUrl,
    directUser: directUser,
    previousBatch: previousBatch,
    creator: creator
  );
}
